@@ -24,6 +24,7 @@ import java.security.PrivilegedExceptionAction;
 import java.text.MessageFormat;
 import java.util.ArrayList;
 import java.util.Arrays;
+import java.util.Collections;
 import java.util.Dictionary;
 import java.util.Enumeration;
 import java.util.HashMap;
@@ -36,6 +37,7 @@ import java.util.Map;
 import java.util.Map.Entry;
 import java.util.ResourceBundle;
 import java.util.Set;
+import java.util.concurrent.ConcurrentSkipListSet;
 
 import javax.servlet.GenericServlet;
 import javax.servlet.ServletConfig;
@@ -48,9 +50,9 @@ import javax.servlet.http.HttpServletRequest;
 import javax.servlet.http.HttpServletRequestWrapper;
 import javax.servlet.http.HttpServletResponse;
 
-import org.apache.commons.io.FilenameUtils;
 import org.apache.felix.webconsole.AbstractWebConsolePlugin;
 import org.apache.felix.webconsole.BrandingPlugin;
+import org.apache.felix.webconsole.User;
 import org.apache.felix.webconsole.WebConsoleConstants;
 import org.apache.felix.webconsole.WebConsoleSecurityProvider;
 import org.apache.felix.webconsole.WebConsoleSecurityProvider2;
@@ -73,6 +75,7 @@ import org.osgi.service.http.HttpContext;
 import org.osgi.service.http.HttpService;
 import org.osgi.service.log.LogService;
 import org.osgi.util.tracker.ServiceTracker;
+import org.osgi.util.tracker.ServiceTrackerCustomizer;
 
 /**
  * The <code>OSGi Manager</code> is the actual Web Console Servlet which
@@ -92,6 +95,7 @@ public class OsgiManager extends GenericServlet
      *
      * @deprecated use {@link WebConsoleConstants#ATTR_APP_ROOT} instead
      */
+    @Deprecated
     private static final String ATTR_APP_ROOT_OLD = OsgiManager.class.getName()
         + ".appRoot";
 
@@ -102,6 +106,7 @@ public class OsgiManager extends GenericServlet
      *
      * @deprecated use {@link WebConsoleConstants#ATTR_LABEL_MAP} instead
      */
+    @Deprecated
     private static final String ATTR_LABEL_MAP_OLD = OsgiManager.class.getName()
         + ".labelMap";
 
@@ -136,6 +141,10 @@ public class OsgiManager extends GenericServlet
 
     private static final String FRAMEWORK_PROP_LOCALE = "felix.webconsole.locale"; //$NON-NLS-1$
 
+    static final String FRAMEWORK_PROP_SECURITY_PROVIDERS = "felix.webconsole.security.providers"; //$NON-NLS-1$
+
+    static final String SECURITY_PROVIDER_PROPERTY_NAME = "webconsole.security.provider.id"; //$NON-NLS-1$
+
     static final String PROP_MANAGER_ROOT = "manager.root"; //$NON-NLS-1$
 
     static final String PROP_DEFAULT_RENDER = "default.render"; //$NON-NLS-1$
@@ -154,6 +163,8 @@ public class OsgiManager extends GenericServlet
 
     static final String PROP_LOCALE = "locale"; //$NON-NLS-1$
 
+    static final String PROP_ENABLE_SECRET_HEURISTIC = "secret.heuristic.enabled"; //$NON-NLS-1$
+
     static final String PROP_HTTP_SERVICE_SELECTOR = "http.service.filter"; //$NON-NLS-1$
 
     public static final int DEFAULT_LOG_LEVEL = LogService.LOG_WARNING;
@@ -170,6 +181,9 @@ public class OsgiManager extends GenericServlet
 
     static final String DEFAULT_HTTP_SERVICE_SELECTOR = ""; //$NON-NLS-1$
 
+    /** Default value for secret heuristics */
+    public static final boolean DEFAULT_ENABLE_SECRET_HEURISTIC = false;
+
     private static final String HEADER_AUTHORIZATION = "Authorization"; //$NON-NLS-1$
 
     private static final String HEADER_WWW_AUTHENTICATE = "WWW-Authenticate"; //$NON-NLS-1$
@@ -188,6 +202,8 @@ public class OsgiManager extends GenericServlet
             "org.apache.felix.webconsole.internal.compendium.PreferencesConfigurationPrinter", //$NON-NLS-1$
             "org.apache.felix.webconsole.internal.compendium.WireAdminConfigurationPrinter", //$NON-NLS-1$
             "org.apache.felix.webconsole.internal.core.BundlesConfigurationPrinter", //$NON-NLS-1$
+            "org.apache.felix.webconsole.internal.core.CapabilitiesPrinter", //$NON-NLS-1$
+            "org.apache.felix.webconsole.internal.core.FrameworkPropertiesPrinter", //$NON-NLS-1$
             "org.apache.felix.webconsole.internal.core.PermissionsConfigurationPrinter", //$NON-NLS-1$
             "org.apache.felix.webconsole.internal.core.ServicesConfigurationPrinter", //$NON-NLS-1$
             "org.apache.felix.webconsole.internal.misc.SystemPropertiesPrinter", //$NON-NLS-1$
@@ -202,26 +218,32 @@ public class OsgiManager extends GenericServlet
             "org.apache.felix.webconsole.internal.system.VMStatPlugin", "vmstat", //$NON-NLS-1$ //$NON-NLS-2$
     };
 
+    /** Flag to control whether secret heuristics is enabled */
+    public static volatile boolean ENABLE_SECRET_HEURISTICS = OsgiManager.DEFAULT_ENABLE_SECRET_HEURISTIC;
+
     private BundleContext bundleContext;
 
     private HttpServiceTracker httpServiceTracker;
 
-    private HttpService httpService;
+    private volatile HttpService httpService;
 
     private PluginHolder holder;
 
     private ServiceTracker brandingTracker;
 
-    private ServiceTracker securityProviderTracker;
+    private ServiceTracker<WebConsoleSecurityProvider, WebConsoleSecurityProvider> securityProviderTracker;
 
     private ServiceRegistration configurationListener;
 
     // list of OsgiManagerPlugin instances activated during init. All these
     // instances will have to be deactivated during destroy
-    private List osgiManagerPlugins = new ArrayList();
+    private List<OsgiManagerPlugin> osgiManagerPlugins = new ArrayList<>();
 
     private String webManagerRoot;
 
+    // not-null when the BasicWebConsoleSecurityProvider service is registered
+    private ServiceRegistration<WebConsoleSecurityProvider> basicSecurityServiceRegistration;
+
     // true if the OsgiManager is registered as a Servlet with the HttpService
     private boolean httpServletRegistered;
 
@@ -229,15 +251,19 @@ public class OsgiManager extends GenericServlet
     private boolean httpResourcesRegistered;
 
     // default configuration from framework properties
-    private HashMap defaultConfiguration;
+    private Map<String, Object> defaultConfiguration;
 
     // configuration from Configuration Admin
-    private HashMap configuration;
+    private volatile Map<String, Object> configuration;
 
     // See https://issues.apache.org/jira/browse/FELIX-2267
     private Locale configuredLocale;
 
-    private Set enabledPlugins;
+    private Set<String> enabledPlugins;
+
+    final ConcurrentSkipListSet<String> registeredSecurityProviders = new ConcurrentSkipListSet<String>();
+
+    final Set<String> requiredSecurityProviders;
 
     ResourceBundleManager resourceBundleManager;
 
@@ -266,13 +292,14 @@ public class OsgiManager extends GenericServlet
 
             try
             {
-                Class pluginClass = classLoader.loadClass(pluginClassName);
-                Object plugin = pluginClass.newInstance();
+                final Class<?> pluginClass = classLoader.loadClass(pluginClassName);
+                final Object plugin = pluginClass.getDeclaredConstructor().newInstance();
 
                 if (plugin instanceof OsgiManagerPlugin)
                 {
-                    ((OsgiManagerPlugin) plugin).activate(bundleContext);
-                    osgiManagerPlugins.add(plugin);
+                    final OsgiManagerPlugin p = (OsgiManagerPlugin)plugin;
+                    p.activate(bundleContext);
+                    osgiManagerPlugins.add(p);
                 }
                 if (plugin instanceof BrandingPlugin)
                 {
@@ -318,13 +345,15 @@ public class OsgiManager extends GenericServlet
         brandingTracker = new BrandingServiceTracker(this);
         brandingTracker.open();
 
+        this.requiredSecurityProviders = splitCommaSeparatedString(bundleContext.getProperty(FRAMEWORK_PROP_SECURITY_PROVIDERS));
+
         // add support for pluggable security
-        securityProviderTracker = new ServiceTracker(bundleContext,
-            WebConsoleSecurityProvider.class.getName(), null);
+        securityProviderTracker = new ServiceTracker<>(bundleContext, WebConsoleSecurityProvider.class,
+                                          new UpdateDependenciesStateCustomizer());
         securityProviderTracker.open();
 
         // load the default configuration from the framework
-        this.defaultConfiguration = new HashMap();
+        this.defaultConfiguration = new HashMap<>();
         this.defaultConfiguration.put( PROP_MANAGER_ROOT,
             ConfigurationUtil.getProperty( bundleContext, FRAMEWORK_PROP_MANAGER_ROOT, DEFAULT_MANAGER_ROOT ) );
         this.defaultConfiguration.put( PROP_REALM,
@@ -334,7 +363,7 @@ public class OsgiManager extends GenericServlet
         this.defaultConfiguration.put( PROP_PASSWORD,
             ConfigurationUtil.getProperty( bundleContext, FRAMEWORK_PROP_PASSWORD, DEFAULT_PASSWORD ) );
         this.defaultConfiguration.put( PROP_LOG_LEVEL,
-            new Integer( ConfigurationUtil.getProperty( bundleContext, FRAMEWORK_PROP_LOG_LEVEL, DEFAULT_LOG_LEVEL ) ) );
+            ConfigurationUtil.getProperty( bundleContext, FRAMEWORK_PROP_LOG_LEVEL, DEFAULT_LOG_LEVEL ) );
         this.defaultConfiguration.put( PROP_LOCALE,
             ConfigurationUtil.getProperty( bundleContext, FRAMEWORK_PROP_LOCALE, null ) );
 
@@ -345,6 +374,7 @@ public class OsgiManager extends GenericServlet
         this.configurationListener = bundleContext.registerService( "org.osgi.service.cm.ManagedService", //$NON-NLS-1$
             new ServiceFactory()
             {
+                @Override
                 public Object getService( Bundle bundle, ServiceRegistration registration )
                 {
                     /*
@@ -368,11 +398,12 @@ public class OsgiManager extends GenericServlet
                 }
 
 
+                @Override
                 public void ungetService( Bundle bundle, ServiceRegistration registration, Object service )
                 {
                     // do nothing
                 }
-            }, new Hashtable()
+            }, new Hashtable<String, Object>()
             {
                 {
                     put( Constants.SERVICE_VENDOR, "The Apache Software Foundation" ); //$NON-NLS-1$
@@ -382,6 +413,21 @@ public class OsgiManager extends GenericServlet
             } );
     }
 
+    void updateRegistrationState() {
+        if (this.httpService != null) {
+            if (this.registeredSecurityProviders.containsAll(this.requiredSecurityProviders)) {
+                // register HTTP service
+                registerHttpService();
+                return;
+            } else {
+                log(LogService.LOG_INFO, "Not all requirements met for the Web Console. Required security providers: "
+                        + this.registeredSecurityProviders + " Registered security providers: " + this.registeredSecurityProviders);
+            }
+        }
+        // Not all requirements met, unregister service.
+        unregisterHttpService();
+    }
+
     public void dispose()
     {
         // dispose off held plugins
@@ -402,10 +448,10 @@ public class OsgiManager extends GenericServlet
         }
 
         // deactivate any remaining plugins
-        for (Iterator pi = osgiManagerPlugins.iterator(); pi.hasNext();)
+        for (Iterator<OsgiManagerPlugin> pi = osgiManagerPlugins.iterator(); pi.hasNext();)
         {
-            Object plugin = pi.next();
-            ((OsgiManagerPlugin) plugin).deactivate();
+            OsgiManagerPlugin plugin = pi.next();
+            plugin.deactivate();
         }
 
         // simply remove all operations, we should not be used anymore
@@ -440,6 +486,7 @@ public class OsgiManager extends GenericServlet
     /**
      * @see javax.servlet.GenericServlet#init()
      */
+    @Override
     public void init()
     {
         // base class initialization not needed, since the GenericServlet.init
@@ -452,14 +499,16 @@ public class OsgiManager extends GenericServlet
     /**
      * @see javax.servlet.GenericServlet#service(javax.servlet.ServletRequest, javax.servlet.ServletResponse)
      */
+    @Override
     public void service(final ServletRequest req, final ServletResponse res)
         throws ServletException, IOException
     {
         // don't really expect to be called within a non-HTTP environment
         try
         {
-            AccessController.doPrivileged(new PrivilegedExceptionAction()
+            AccessController.doPrivileged(new PrivilegedExceptionAction<Object>()
             {
+                @Override
                 public Object run() throws Exception
                 {
                     service((HttpServletRequest) req, (HttpServletResponse) res);
@@ -483,10 +532,6 @@ public class OsgiManager extends GenericServlet
                 throw new IOException(x.toString());
             }
         }
-
-        // ensure response has been sent back and response is committed
-        // (we are authorative for our URL space and no other servlet should interfere)
-        res.flushBuffer();
     }
 
     private void ensureLocaleCookieSet(HttpServletRequest request, HttpServletResponse response, Locale locale) {
@@ -519,8 +564,8 @@ public class OsgiManager extends GenericServlet
                 path = path.concat("/"); //$NON-NLS-1$
             }
             path = path.concat(holder.getDefaultPluginLabel());
-            response.sendRedirect(path);
             response.setContentLength(0);
+            response.sendRedirect(path);
             return;
         }
 
@@ -623,7 +668,7 @@ public class OsgiManager extends GenericServlet
             // check for basic authentication
             String auth = request.getHeader(HEADER_AUTHORIZATION); //$NON-NLS-1$
             if (null != auth && auth.toLowerCase().startsWith("basic ")) { //$NON-NLS-1$
-                Map config = getConfiguration();
+                Map<String, Object> config = getConfiguration();
                 String realm = ConfigurationUtil.getProperty(config, PROP_REALM, DEFAULT_REALM);
                 response.setHeader(HEADER_WWW_AUTHENTICATE, "Basic realm=\"" +  realm + "\""); //$NON-NLS-1$ //$NON-NLS-2$
                 response.addCookie(new Cookie("logout", "true")); //$NON-NLS-1$ //$NON-NLS-2$
@@ -635,6 +680,7 @@ public class OsgiManager extends GenericServlet
         request.removeAttribute(HttpContext.REMOTE_USER);
         request.removeAttribute(HttpContext.AUTHORIZATION);
         request.removeAttribute(WebConsoleSecurityProvider2.USER_ATTRIBUTE);
+        request.removeAttribute(User.USER_ATTRIBUTE);
     }
 
     private final AbstractWebConsolePlugin getConsolePlugin(final String label)
@@ -681,6 +727,10 @@ public class OsgiManager extends GenericServlet
             locale = configuredLocale;
         if (locale == null)
             locale = request.getLocale();
+        // this should never happen as request.getLocale()
+        // must return a locale (and not null). But just as a sanity check.
+        if (locale == null)
+            locale = Locale.ENGLISH;
 
         return locale;
     }
@@ -688,6 +738,7 @@ public class OsgiManager extends GenericServlet
     /**
      * @see javax.servlet.GenericServlet#destroy()
      */
+    @Override
     public void destroy()
     {
         // base class destroy not needed, since the GenericServlet.destroy
@@ -795,6 +846,7 @@ public class OsgiManager extends GenericServlet
             /**
              * @see javax.servlet.ServletRequestWrapper#getLocale()
              */
+            @Override
             public Locale getLocale()
             {
                 return locale;
@@ -811,11 +863,8 @@ public class OsgiManager extends GenericServlet
         return new FilteringResponseWrapper(response, resourceBundle, request);
     }
 
-    private static class HttpServiceTracker extends ServiceTracker
+    private static class HttpServiceTracker extends ServiceTracker<HttpService, HttpService>
     {
-
-        private static final String HTTP_SERVICE = "org.osgi.service.http.HttpService"; //$NON-NLS-1$
-
         private final OsgiManager osgiManager;
 
         private final String httpServiceSelector;
@@ -829,7 +878,7 @@ public class OsgiManager extends GenericServlet
                 try
                 {
                     final String filterString = "(&(" + Constants.OBJECTCLASS + "=" //$NON-NLS-1$ //$NON-NLS-2$
-                        + HTTP_SERVICE + ")(" + httpServiceSelector + "))"; //$NON-NLS-1$ //$NON-NLS-2$
+                        + HttpService.class.getName() + ")(" + httpServiceSelector + "))"; //$NON-NLS-1$ //$NON-NLS-2$
                     Filter filter = osgiManager.getBundleContext().createFilter(
                         filterString);
                     return new HttpServiceTracker(osgiManager, httpServiceSelector,
@@ -847,7 +896,7 @@ public class OsgiManager extends GenericServlet
 
         private HttpServiceTracker(final OsgiManager osgiManager)
         {
-            super(osgiManager.getBundleContext(), HTTP_SERVICE, null);
+            super(osgiManager.getBundleContext(), HttpService.class, null);
             this.osgiManager = osgiManager;
             this.httpServiceSelector = null;
         }
@@ -868,56 +917,48 @@ public class OsgiManager extends GenericServlet
             return httpServiceSelector == null;
         }
 
-        public Object addingService(ServiceReference reference)
+        @Override
+        public HttpService addingService(ServiceReference<HttpService> reference)
         {
-            Object service = super.addingService(reference);
-            if (service instanceof HttpService)
-            {
-                osgiManager.bindHttpService((HttpService) service);
-            }
+            HttpService service = super.addingService(reference);
+            osgiManager.bindHttpService(service);
             return service;
         }
 
-        public void removedService(ServiceReference reference, Object service)
+        @Override
+        public void removedService(ServiceReference<HttpService> reference, HttpService service)
         {
-            if (service instanceof HttpService)
-            {
-                osgiManager.unbindHttpService((HttpService) service);
-            }
+            osgiManager.unbindHttpService(service);
 
             super.removedService(reference, service);
         }
     }
 
-    private static class BrandingServiceTracker extends ServiceTracker
+    private static class BrandingServiceTracker extends ServiceTracker<BrandingPlugin, BrandingPlugin>
     {
         BrandingServiceTracker(OsgiManager osgiManager)
         {
-            super(osgiManager.getBundleContext(), BrandingPlugin.class.getName(), null);
+            super(osgiManager.getBundleContext(), BrandingPlugin.class, null);
         }
 
-        public Object addingService(ServiceReference reference)
+        @Override
+        public BrandingPlugin addingService(ServiceReference<BrandingPlugin> reference)
         {
-            Object plugin = super.addingService(reference);
-            if (plugin instanceof BrandingPlugin)
-            {
-                AbstractWebConsolePlugin.setBrandingPlugin((BrandingPlugin) plugin);
-            }
+            BrandingPlugin plugin = super.addingService(reference);
+            AbstractWebConsolePlugin.setBrandingPlugin( plugin);
             return plugin;
         }
 
-        public void removedService(ServiceReference reference, Object service)
+        @Override
+        public void removedService(ServiceReference<BrandingPlugin> reference, BrandingPlugin service)
         {
-            if (service instanceof BrandingPlugin)
-            {
-                AbstractWebConsolePlugin.setBrandingPlugin(null);
-            }
+            AbstractWebConsolePlugin.setBrandingPlugin(null);
             super.removedService(reference, service);
         }
 
     }
 
-    protected synchronized void bindHttpService(HttpService httpService)
+    protected void bindHttpService(HttpService httpService)
     {
         // do not bind service, when we are already bound
         if (this.httpService != null)
@@ -927,7 +968,12 @@ public class OsgiManager extends GenericServlet
             return;
         }
 
-        Map config = getConfiguration();
+        this.httpService = httpService;
+        updateRegistrationState();
+    }
+
+    synchronized void registerHttpService() {
+        Map<String, Object> config = getConfiguration();
 
         // get authentication details
         String realm = ConfigurationUtil.getProperty(config, PROP_REALM, DEFAULT_REALM);
@@ -938,30 +984,42 @@ public class OsgiManager extends GenericServlet
         try
         {
             HttpContext httpContext = new OsgiManagerHttpContext(httpService,
-                securityProviderTracker, userId, password, realm);
+                securityProviderTracker, realm);
 
-            Dictionary servletConfig = toStringConfig(config);
+            Dictionary<String, String> servletConfig = toStringConfig(config);
 
-            // register this servlet and take note of this
-            httpService.registerServlet(this.webManagerRoot, this, servletConfig,
-                httpContext);
-            httpServletRegistered = true;
+            if (basicSecurityServiceRegistration == null) {
+                //register this component
+                BasicWebConsoleSecurityProvider service = new BasicWebConsoleSecurityProvider(bundleContext,
+                        userId, password, realm);
+                Dictionary<String, Object> serviceProperties = new Hashtable<>(); // NOSONAR
+                // this is a last resort service, so use a low service ranking to prefer all other services over this one
+                serviceProperties.put(Constants.SERVICE_RANKING, Integer.MIN_VALUE);
+                basicSecurityServiceRegistration = bundleContext.registerService(WebConsoleSecurityProvider.class,
+                        service, serviceProperties);
+            }
 
-            // register resources and take of this
-            httpService.registerResources(this.webManagerRoot + "/res", "/res",
-                httpContext);
-            httpResourcesRegistered = true;
+            if (!httpServletRegistered) {
+                // register this servlet and take note of this
+                httpService.registerServlet(this.webManagerRoot, this, servletConfig,
+                    httpContext);
+                httpServletRegistered = true;
+            }
 
+            if (!httpResourcesRegistered) {
+                // register resources and take of this
+                httpService.registerResources(this.webManagerRoot + "/res", "/res",
+                    httpContext);
+                httpResourcesRegistered = true;
+            }
         }
         catch (Exception e)
         {
             log(LogService.LOG_ERROR, "bindHttpService: Problem setting up", e);
         }
-
-        this.httpService = httpService;
     }
 
-    protected synchronized void unbindHttpService(HttpService httpService)
+    protected void unbindHttpService(HttpService httpService)
     {
         if (this.httpService != httpService)
         {
@@ -970,8 +1028,25 @@ public class OsgiManager extends GenericServlet
             return;
         }
 
+        unregisterHttpService();
+
         // drop the service reference
         this.httpService = null;
+    }
+
+    synchronized void unregisterHttpService() {
+        if (httpService == null)
+            return;
+
+        if (basicSecurityServiceRegistration != null) {
+            try {
+                basicSecurityServiceRegistration.unregister();
+            } catch (Throwable t) {
+                log(LogService.LOG_WARNING,
+                        "unbindHttpService: Failed unregistering basic WebConsoleSecurityProvider", t);
+            }
+            basicSecurityServiceRegistration = null;
+        }
 
         if (httpResourcesRegistered)
         {
@@ -1003,27 +1078,27 @@ public class OsgiManager extends GenericServlet
     }
 
 
-    private Map getConfiguration()
+    private Map<String, Object> getConfiguration()
     {
         return configuration;
     }
 
 
-    Map getDefaultConfiguration()
+    Map<String, Object> getDefaultConfiguration()
     {
         return defaultConfiguration;
     }
 
 
-    synchronized void updateConfiguration( Dictionary osgiConfig )
+    synchronized void updateConfiguration( Dictionary<String, Object> osgiConfig )
     {
-        HashMap config = new HashMap( this.defaultConfiguration );
+        Map<String, Object> config = new HashMap<String, Object>( this.defaultConfiguration );
 
         if ( osgiConfig != null )
         {
-            for ( Enumeration keys = osgiConfig.keys(); keys.hasMoreElements(); )
+            for ( Enumeration<String> keys = osgiConfig.keys(); keys.hasMoreElements(); )
             {
-                final Object key = keys.nextElement();
+                final String key = keys.nextElement();
                 config.put( key, osgiConfig.get( key ) );
             }
         }
@@ -1061,9 +1136,13 @@ public class OsgiManager extends GenericServlet
             httpServiceTracker = null;
         }
 
+        // secret heuristics
+        final boolean enableHeuristics = ConfigurationUtil.getProperty(config, PROP_ENABLE_SECRET_HEURISTIC, DEFAULT_ENABLE_SECRET_HEURISTIC);
+        OsgiManager.ENABLE_SECRET_HEURISTICS = enableHeuristics;
+
         // get enabled plugins
         String[] plugins = ConfigurationUtil.getStringArrayProperty(config, PROP_ENABLED_PLUGINS);
-        enabledPlugins = null == plugins ? null : new HashSet(Arrays.asList(plugins));
+        enabledPlugins = null == plugins ? null : new HashSet<String>(Arrays.asList(plugins));
         // check for moved config manager class (see FELIX-4074)
         if ( enabledPlugins != null )
         {
@@ -1138,43 +1217,86 @@ public class OsgiManager extends GenericServlet
     }
 
 
-    private Dictionary toStringConfig( Map config )
+    private Dictionary<String, String> toStringConfig( Map<String, Object> config )
     {
-        Dictionary stringConfig = new Hashtable();
-        for ( Iterator ei = config.entrySet().iterator(); ei.hasNext(); )
+        Dictionary<String, String> stringConfig = new Hashtable<>();
+        for ( Iterator<Map.Entry<String, Object>> ei = config.entrySet().iterator(); ei.hasNext(); )
         {
-            Entry entry = ( Entry ) ei.next();
+            Entry<String, Object> entry = ei.next();
             stringConfig.put( entry.getKey(), String.valueOf( entry.getValue() ) );
         }
         return stringConfig;
     }
 
-    private Map langMap;
+    static Set<String> splitCommaSeparatedString(final String str) {
+        if (str == null)
+            return Collections.emptySet();
+
+        final Set<String> values = new HashSet<String>();
+        for (final String s : str.split(",")) {
+            String trimmed = s.trim();
+            if (trimmed.length() > 0) {
+                values.add(trimmed);
+            }
+        }
+        return Collections.unmodifiableSet(values);
+    }
+
+    private Map<String, String> langMap;
 
 
-    private final Map getLangMap()
+    private final Map<String, String> getLangMap()
     {
         if (null != langMap)
             return langMap;
-        final Map map = new HashMap();
+        final Map<String, String> map = new HashMap<>();
         final Bundle bundle = bundleContext.getBundle();
-        final Enumeration e = bundle.findEntries("res/flags", null, false); //$NON-NLS-1$
+        final Enumeration<URL> e = bundle.findEntries("res/flags", null, false); //$NON-NLS-1$
         while (e != null && e.hasMoreElements())
         {
-            final URL img = (URL) e.nextElement();
-            final String name = FilenameUtils.getBaseName(img.getFile());
-            try
-            {
+            final URL img = e.nextElement();
+            final String path = img.getPath();
+            try {
+                final int lastSlash = path.lastIndexOf('/');
+                final int dot = path.indexOf('.', lastSlash);
+                final String name = (dot == -1 ? path.substring(lastSlash+1) : path.substring(lastSlash + 1, dot));
                 final String locale = new Locale(name, "").getDisplayLanguage(); //$NON-NLS-1$
                 map.put(name, null != locale ? locale : name);
             }
-            catch (Throwable t)
-            {
-                t.printStackTrace();
-                /* ignore invalid locale? */
+            catch (Throwable t) {
+                // Ignore invalid locale?
             }
         }
         return langMap = map;
     }
 
+    class UpdateDependenciesStateCustomizer implements ServiceTrackerCustomizer<WebConsoleSecurityProvider, WebConsoleSecurityProvider> {
+        @Override
+        public WebConsoleSecurityProvider addingService(ServiceReference<WebConsoleSecurityProvider> reference) {
+            Object nameObj = reference.getProperty(SECURITY_PROVIDER_PROPERTY_NAME);
+            if (nameObj instanceof String) {
+                String name = (String) nameObj;
+                registeredSecurityProviders.add(name);
+                updateRegistrationState();
+            }
+            return bundleContext.getService(reference);
+        }
+
+        @Override
+        public void modifiedService(ServiceReference<WebConsoleSecurityProvider> reference, WebConsoleSecurityProvider service) {
+            removedService(reference, service);
+            addingService(reference);
+        }
+
+        @Override
+        public void removedService(ServiceReference<WebConsoleSecurityProvider> reference, WebConsoleSecurityProvider service) {
+            Object nameObj = reference.getProperty(SECURITY_PROVIDER_PROPERTY_NAME);
+            if (nameObj instanceof String) {
+                String name = (String) nameObj;
+                registeredSecurityProviders.remove(name);
+                updateRegistrationState();
+            }
+        }
+
+    }
 }
